package sun.awt.im;

import java.security.AccessController;
import sun.security.action.GetPropertyAction;
import java.awt.im.spi.InputMethod;
import java.awt.HeadlessException;
import sun.awt.InputMethodSupport;
import java.awt.Toolkit;
import java.awt.GraphicsEnvironment;
import javax.swing.JFrame;
import java.awt.Window;
import java.awt.im.InputMethodRequests;
import java.awt.Rectangle;
import java.awt.AWTEvent;
import java.text.AttributedString;
import java.awt.event.KeyEvent;
import java.awt.event.InputMethodEvent;
import java.awt.font.TextHitInfo;
import java.text.AttributedCharacterIterator;
import java.awt.Component;

public class InputMethodContext extends InputContext implements java.awt.im.spi.InputMethodContext
{
    private boolean dispatchingCommittedText;
    private CompositionAreaHandler compositionAreaHandler;
    private Object compositionAreaHandlerLock;
    private static boolean belowTheSpotInputRequested;
    private boolean inputMethodSupportsBelowTheSpot;
    
    public InputMethodContext() {
        this.compositionAreaHandlerLock = new Object();
    }
    
    void setInputMethodSupportsBelowTheSpot(boolean inputMethodSupportsBelowTheSpot) {
        this.inputMethodSupportsBelowTheSpot = inputMethodSupportsBelowTheSpot;
    }
    
    boolean useBelowTheSpotInput() {
        return InputMethodContext.belowTheSpotInputRequested && this.inputMethodSupportsBelowTheSpot;
    }
    
    private boolean haveActiveClient() {
        Component clientComponent = this.getClientComponent();
        return clientComponent != null && clientComponent.getInputMethodRequests() != null;
    }
    
    @Override
    public void dispatchInputMethodEvent(int id, AttributedCharacterIterator text, int committedCharacterCount, TextHitInfo caret, TextHitInfo visiblePosition) {
        Component clientComponent = this.getClientComponent();
        if (clientComponent != null) {
            InputMethodEvent e = new InputMethodEvent(clientComponent, id, text, committedCharacterCount, caret, visiblePosition);
            if (this.haveActiveClient() && !this.useBelowTheSpotInput()) {
                clientComponent.dispatchEvent(e);
            }
            else {
                this.getCompositionAreaHandler(true).processInputMethodEvent(e);
            }
        }
    }
    
    synchronized void dispatchCommittedText(Component component, AttributedCharacterIterator text, int committedCharacterCount) {
        if (committedCharacterCount == 0 || text.getEndIndex() <= text.getBeginIndex()) {
            return;
        }
        long currentTimeMillis = System.currentTimeMillis();
        this.dispatchingCommittedText = true;
        try {
            if (component.getInputMethodRequests() != null) {
                int beginIndex = text.getBeginIndex();
                component.dispatchEvent(new InputMethodEvent(component, 1100, new AttributedString(text, beginIndex, beginIndex + committedCharacterCount).getIterator(), committedCharacterCount, null, null));
            }
            else {
                for (int keyChar = text.first(); committedCharacterCount-- > 0 && keyChar != '\uffff'; keyChar = text.next()) {
                    component.dispatchEvent(new KeyEvent(component, 400, currentTimeMillis, 0, 0, (char)keyChar));
                }
            }
        }
        finally {
            this.dispatchingCommittedText = false;
        }
        return;
    }
    
    @Override
    public void dispatchEvent(AWTEvent awtEvent) {
        if ((awtEvent instanceof InputMethodEvent)) {
            if (((Component)awtEvent.getSource()).getInputMethodRequests() == null || (this.useBelowTheSpotInput() && !this.dispatchingCommittedText)) {
                this.getCompositionAreaHandler(true).processInputMethodEvent((InputMethodEvent)awtEvent);
            }
        }
        else if (!this.dispatchingCommittedText) {
            super.dispatchEvent(awtEvent);
        }
    }
    
    private CompositionAreaHandler getCompositionAreaHandler(boolean b) {
        synchronized (this.compositionAreaHandlerLock) {
            if (this.compositionAreaHandler == null) {
                this.compositionAreaHandler = new CompositionAreaHandler(this);
            }
            this.compositionAreaHandler.setClientComponent(this.getClientComponent());
            if (b) {
                this.compositionAreaHandler.grabCompositionArea(false);
            }
            return this.compositionAreaHandler;
        }
    }
    
    void grabCompositionArea(boolean b) {
        synchronized (this.compositionAreaHandlerLock) {
            if (this.compositionAreaHandler != null) {
                this.compositionAreaHandler.grabCompositionArea(b);
            }
            else {
                CompositionAreaHandler.closeCompositionArea();
            }
        }
    }
    
    void releaseCompositionArea() {
        synchronized (this.compositionAreaHandlerLock) {
            if (this.compositionAreaHandler != null) {
                this.compositionAreaHandler.releaseCompositionArea();
            }
        }
    }
    
    boolean isCompositionAreaVisible() {
        return this.compositionAreaHandler != null && this.compositionAreaHandler.isCompositionAreaVisible();
    }
    
    void setCompositionAreaVisible(boolean compositionAreaVisible) {
        if (this.compositionAreaHandler != null) {
            this.compositionAreaHandler.setCompositionAreaVisible(compositionAreaVisible);
        }
    }
    
    @Override
    public Rectangle getTextLocation(TextHitInfo textHitInfo) {
        return this.getReq().getTextLocation(textHitInfo);
    }
    
    @Override
    public TextHitInfo getLocationOffset(int n, int n2) {
        return this.getReq().getLocationOffset(n, n2);
    }
    
    @Override
    public int getInsertPositionOffset() {
        return this.getReq().getInsertPositionOffset();
    }
    
    @Override
    public AttributedCharacterIterator getCommittedText(int n, int n2, AttributedCharacterIterator.Attribute[] array) {
        return this.getReq().getCommittedText(n, n2, array);
    }
    
    @Override
    public int getCommittedTextLength() {
        return this.getReq().getCommittedTextLength();
    }
    
    @Override
    public AttributedCharacterIterator cancelLatestCommittedText(AttributedCharacterIterator.Attribute[] array) {
        return this.getReq().cancelLatestCommittedText(array);
    }
    
    @Override
    public AttributedCharacterIterator getSelectedText(AttributedCharacterIterator.Attribute[] array) {
        return this.getReq().getSelectedText(array);
    }
    
    private InputMethodRequests getReq() {
        if (this.haveActiveClient() && !this.useBelowTheSpotInput()) {
            return this.getClientComponent().getInputMethodRequests();
        }
        return this.getCompositionAreaHandler(false);
    }
    
    @Override
    public Window createInputMethodWindow(String s, boolean b) {
        return createInputMethodWindow(s, b ? this : null, false);
    }
    
    @Override
    public JFrame createInputMethodJFrame(String s, boolean b) {
        return (JFrame)createInputMethodWindow(s, b ? this : null, true);
    }
    
    static Window createInputMethodWindow(String s, InputContext inputContext, boolean b) {
        if (GraphicsEnvironment.isHeadless()) {
            throw new HeadlessException();
        }
        if (b) {
            return new InputMethodJFrame(s, inputContext);
        }
        Toolkit defaultToolkit = Toolkit.getDefaultToolkit();
        if ((defaultToolkit instanceof InputMethodSupport)) {
            return ((InputMethodSupport)defaultToolkit).createInputMethodWindow(s, inputContext);
        }
        throw new InternalError("Input methods must be supported");
    }
    
    @Override
    public void enableClientWindowNotification(InputMethod inputMethod, boolean b) {
        super.enableClientWindowNotification(inputMethod, b);
    }
    
    void setCompositionAreaUndecorated(boolean compositionAreaUndecorated) {
        if (this.compositionAreaHandler != null) {
            this.compositionAreaHandler.setCompositionAreaUndecorated(compositionAreaUndecorated);
        }
    }
    
    static {
        String property = (String)AccessController.doPrivileged(new GetPropertyAction("java.awt.im.style", null));
        if (property == null) {
            property = Toolkit.getProperty("java.awt.im.style", null);
        }
        InputMethodContext.belowTheSpotInputRequested = "below-the-spot".equals(property);
    }
}
